(*
  Copyright (C) BitBlaze, 2009-2010. 

#Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.

#THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*)
(**
  A classifier for x86 instructions
  Includes other generally useful functions to operate with traces such as 
    obtaining the list of operands in an instruction or checking if an 
    instruction is tainted
  The x86 classifier can be considered a very basic disassembler for 
    some x86 instructions. 
  It should be replaced by a proper disassembler some time in the future 

*)

(** Instruction prefixes *)
type prefixType =
  | ADSIZE
  | OPSIZE
  | REPE
  | REPNE
  | LOCK
  | SEG_CS
  | SEG_DS
  | SEG_ES
  | SEG_FS
  | SEG_GS
  | SEG_SS

(** Instruction mnemonics *)
type insnMnemonic =
  (* Special mnemonics *)
  | UNSUPPORTED (** Instruction not disassembled *)
  | UNASSIGNED (** Opcode does not exist, has not been assigned *)
  | ERROR (** Found incorrect opcode *)
  | FLOAT (** Floating point opcode, individual opcodes not extracted yet *)
  (* Instruction mnemonics *)
  | M_AAA
  | M_AAD
  | M_AAM
  | M_AAS
  | M_ADC
  | M_ADD
  | M_AND
  | M_ARPL
  | M_BOUND
  | M_BSWAP
  | M_BTR
  | M_CALL
  | M_CBW
  | M_CLC
  | M_CLD
  | M_CLI
  | M_CMC
  | M_CMP
  | M_CMPS
  | M_CMPXCHG
  | M_CPUID
  | M_CWD
  | M_DAA
  | M_DAS
  | M_DEC
  | M_DIV
  | M_ENTER
  | M_HLT
  | M_ICALL
  | M_ICEBP
  | M_IDIV
  | M_IJMP
  | M_IMUL
  | M_IN
  | M_INC
  | M_INS
  | M_INT
  | M_INT3
  | M_INTO
  | M_IRET
  | M_JA
  | M_JAE
  | M_JB
  | M_JBE
  | M_JE
  | M_JECX
  | M_JG
  | M_JGE
  | M_JL
  | M_JLE
  | M_JMP
  | M_JNB
  | M_JNBE
  | M_JNE
  | M_JNL
  | M_JNLE
  | M_JNO
  | M_JNP
  | M_JNS
  | M_JNZ
  | M_JO
  | M_JP
  | M_JS
  | M_JZ
  | M_LAHF
  | M_LDS
  | M_LEA
  | M_LEAVE
  | M_LES
  | M_LFS
  | M_LGS
  | M_LODS
  | M_LOOP
  | M_LOOPE
  | M_LOOPNE
  | M_LSS
  | M_MOV
  | M_MOVD
  | M_MOVQ
  | M_MOVS
  | M_MOVZX
  | M_MUL
  | M_NEG
  | M_NOP
  | M_NOT
  | M_OR
  | M_OUT
  | M_OUTS
  | M_POP
  | M_POPA
  | M_POPF
  | M_PUSH
  | M_PUSHA
  | M_PUSHF
  | M_RCL
  | M_RCR
  | M_RDTSC
  | M_RET
  | M_ROL
  | M_ROR
  | M_SAHF
  | M_SAL
  | M_SALC
  | M_SAR
  | M_SBB
  | M_SCAS
  | M_SHL
  | M_SHR
  | M_STC
  | M_STD
  | M_STI
  | M_STOS
  | M_SUB
  | M_SYSENTER
  | M_TEST
  | M_WAIT
  | M_XCHG
  | M_XLAT
  | M_XOR

(** Disassembly information for an instruction *)
type insnInfo = {
  disas_prefix_l : prefixType list; (** List of prefixes *)
  disas_opcode : int; (** Opcode, after removing prefixes *)
  disas_two_byte_opcode : bool; (** If true opcode is (0x0f,disas_opcode) *)
  disas_maskedmodrm : int; (** Masked modrm *)
}

(** [disas_insn rawbytes] disassembles the instruction and returns an 
  insnInfo record with information *)
val disas_insn : char array -> insnInfo

(** [is_rep_insn insn] returns whether [insn] uses a REPE or REPNE prefix *)
val is_rep_insn : Temu_trace.instruction -> bool

(** [insn_class rawbytes] returns then mnemonic for the instruction with the 
  given [rawbytes] *)
val insn_class : char array -> insnMnemonic

(** [get_operands_in_insn insn] returns the list of operands used by the 
  instruction [insn]
  Adds the memory operands and esp optionally (enabled by default) *)
val get_operands_in_insn : ?include_memregs : bool -> ?include_esp : bool -> 
  Temu_trace.instruction -> Temu_trace.operand list

(* Get list of memory operands, with associated memregs, in an instruction *)
(** [get_mem_operands_in_insn insn] returns the list of memory addressing 
  operands used by the instruction [insn] *)
val get_mem_operands_in_insn : Temu_trace.instruction -> 
  (Temu_trace.operand * Temu_trace.operand array) list

(** [get_rw_operands insn] returns a pair of lists containing the operands 
  read and written by the instruction [insn]. 
  Operands read and written appear on both lists
  Operands conditionally read or written appear in the output lists regardless 
    of whether they were actually read or written by the instruction

  Optionally includes immediates, esp and memory addressing operands, 
    but by default it does not. 
*)
val get_rw_operands : ?include_immediates : bool -> ?include_esp : bool -> 
  ?include_memregs : bool -> Temu_trace.instruction -> 
  Temu_trace.operand list * Temu_trace.operand list

(** [get_op_byteval op idx] returns the int32 value of the byte at 
  index [idx] in operand [op] *)
val get_op_byteval : Temu_trace.operand -> int -> int32

(** [is_op_tainted op] returns whether the operand [op] is tainted *) 
val is_op_tainted : Temu_trace.operand -> bool

(** [byte_tainted op idx] returns whether the byte at offset [idx] in 
  operand [op] is tainted *) 
val byte_tainted : Temu_trace.operand -> int -> bool

(** [is_tainted_instruction insn] returns whether the instruction [insn] 
  is tainted *) 
val is_tainted_instruction : Temu_trace.instruction -> bool

(** [add_taint_info_in_insn_to_list taint_l insn] adds the taint data 
  accessed by the instruction [insn] to the taint list [taint_l]
  The taint data is a pair (origin,offset) and the taint list is a 
  list of such pairs 
  By default, it does not add the taint from memory registers and esp, 
  but this can be changed using the optional parameters *)
val add_taint_info_in_insn_to_list : ?include_memregs : bool -> 
  ?include_esp : bool -> (int32 * int) list -> 
  Temu_trace.instruction -> (int32 * int) list

(** [get_insn_taint insn] returns a list of the taint pairs (origin,offset) 
  accessed by the instruction [insn] 
  Shortcut for add_taint_info_in_insn_to_list with an empty taint list.
  By default, it does not add the taint from memory registers and esp, 
  but this can be changed using the optional parameters *)
val get_insn_taint : ?include_memregs : bool -> ?include_esp : bool -> 
  Temu_trace.instruction -> (int32 * int) list

(** [is_esp_op op] returns whether the operand [op] is the ESP register *) 
val is_esp_op : Temu_trace.operand -> bool

(** [is_ebp_op op] returns whether the operand [op] is the EBP register *) 
val is_ebp_op : Temu_trace.operand -> bool

(** [is_eax_op op] returns whether the operand [op] is the EAX register *) 
val is_eax_op : Temu_trace.operand -> bool

(** [is_ebx_op op] returns whether the operand [op] is the EBX register *) 
val is_ebx_op : Temu_trace.operand -> bool

(** [is_ecx_op op] returns whether the operand [op] is the ECX register *) 
val is_ecx_op : Temu_trace.operand -> bool

(** [is_edx_op op] returns whether the operand [op] is the EDX register *) 
val is_edx_op : Temu_trace.operand -> bool

(** [is_esi_op op] returns whether the operand [op] is the ESI register *) 
val is_esi_op : Temu_trace.operand -> bool

(** [is_edi_op op] returns whether the operand [op] is the EDI register *) 
val is_edi_op : Temu_trace.operand -> bool

(** [is_call_raw itype] returns whether the given instruction mnemonic [itype]
  corresponds to a call instruction *) 
val is_call_raw : insnMnemonic -> bool

(** [is_call insn] returns whether the given instruction [insn] is a 
  call instruction *)
val is_call : Temu_trace.instruction -> bool

(** [is_ret_raw itype] returns whether the given instruction mnemonic [itype]
  corresponds to a return instruction *) 
val is_ret_raw : insnMnemonic -> bool

(** [is_ret insn] returns whether the given instruction [insn] is a 
  return instruction *)
val is_ret : Temu_trace.instruction -> bool

(** [is_xor_identical insn] returns whether the given instruction [insn] is an 
  XOR operation where both operands are the identical. E.g., xor %eax,%eax *)
val is_xor_identical : Temu_trace.instruction -> bool

(** [is_test_identical insn] returns whether the given instruction [insn] is a 
  TEST operation where both operands are the identical. E.g., test %eax,%eax *)
val is_test_identical : Temu_trace.instruction -> bool

(** [is_sub_identical insn] returns whether the given instruction [insn] is a 
  substraction where both operands are the identical. E.g., sub %eax,%eax *)
val is_sub_identical : Temu_trace.instruction -> bool

(** [is_and_with_zero_imm insn] returns whether the given instruction [insn] 
  is an AND operation between some operand and an immediate of value zero *)
val is_and_with_zero_imm : Temu_trace.instruction -> bool

(** [is_or_with_minus_one_imm insn] returns whether the given instruction [insn]
  is an OR operation between some operand and an immediate of value -1. 
  E.g., or %eax,$0xffffffff *)
val is_or_with_minus_one_imm : Temu_trace.instruction -> bool

(** [is_xor_with_minus_one insn] returns whether the given instruction [insn]
  is an XOR operation between some operand and an immediate of value -1. 
  E.g., or %eax,$0xffffffff *)
val is_xor_with_minus_one : Temu_trace.instruction -> bool

(** [is_float_raw itype] returns whether the given instruction mnemonic [itype]
  corresponds to a floating point instruction *) 
val is_float_raw : insnMnemonic -> bool

(** [is_float insn] returns whether the given instruction [insn] is a 
  floating point instruction *)
val is_float : Temu_trace.instruction -> bool

(** [is_preparation_raw itype] returns whether the given instruction mnemonic 
  [itype] corresponds to a instruction that moves data without performing 
  other arithmetic/bitwise operation. E.g., mov,movs,lea,pop,push,stos... *) 
val is_preparation_raw : insnMnemonic -> bool

(** [is_preparation insn] returns whether the given instruction [insn]  
  is a instruction that moves data without performing 
  other arithmetic/bitwise operation. E.g., mov,movs,lea,pop,push,stos... *) 
val is_preparation : Temu_trace.instruction -> bool

(** [is_uncond_jump_raw itype] returns whether the given instruction 
  mnemonic [itype] corresponds to a unconditional jump instruction.
  E.g., jmp (%eax) *) 
val is_uncond_jump_raw : insnMnemonic -> bool

(** [is_uncond_jump insn] returns whether the given instruction [insn] is an 
  unconditional jump instruction *)
val is_uncond_jump : Temu_trace.instruction -> bool

(** [is_cond_jump_raw itype] returns whether the given instruction 
  mnemonic [itype] corresponds to a conditional jump instruction.
  E.g., je 0x0000000071a52c0b  *) 
val is_cond_jump_raw : insnMnemonic -> bool

(** [is_cond_jump insn] returns whether the given instruction [insn] is a 
  conditional jump instruction *)
val is_cond_jump : Temu_trace.instruction -> bool

(** [is_ind_jump_raw itype] returns whether the given instruction 
  mnemonic [itype] corresponds to an indirect jump instruction.
  E.g., jmp *0x41e110  *) 
val is_ind_jump_raw : insnMnemonic -> bool

(** [is_ind_jump insn] returns whether the given instruction [insn] is an 
  indirect jump instruction *)
val is_ind_jump : Temu_trace.instruction -> bool

(** [is_ind_call_raw itype] returns whether the given instruction 
  mnemonic [itype] corresponds to an indirect call instruction.
  E.g., call *0x41e110  *) 
val is_ind_call_raw : insnMnemonic -> bool

(** [is_ind_call insn] returns whether the given instruction [insn] is an 
  indirect call instruction *)
val is_ind_call : Temu_trace.instruction -> bool

(** [is_push_raw itype] returns whether the given instruction mnemonic [itype] 
  corresponds to push/pusha/pushf instruction. E.g., push %ebp *) 
val is_push_raw : insnMnemonic -> bool

(** [is_push insn] returns whether the given instruction [insn] is a 
  push/pusha/pushf instruction *)
val is_push : Temu_trace.instruction -> bool

(** [is_comparison insn] returns whether the given instruction [insn] is a 
  comparison instruction. E.g., cmp/cmps/sub/xor/scas/test (same operands) *)
val is_comparison : Temu_trace.instruction -> bool

(** [is_system_call insn] returns whether the given instruction [insn] is a 
  system call instruction. E.g., systenter/int 0x80/int 0x2e *)
val is_system_call : Temu_trace.instruction -> bool

(** [is_xchg_raw itype] returns whether the given instruction mnemonic [itype] 
  corresponds to an exchange instruction. E.g., xchg %eax,(%esp) *) 
val is_xchg_raw : insnMnemonic -> bool

(** [is_xchg insn] returns whether the given instruction [insn] is an 
  exchange instruction *)
val is_xchg : Temu_trace.instruction -> bool

(** [is_bswap_raw itype] returns whether the given instruction mnemonic [itype] 
  corresponds to a byte swap instruction. E.g., bswap %eax *)
val is_bswap_raw : insnMnemonic -> bool

(** [is_bswap insn] returns whether the given instruction [insn] is a 
  byte swap instruction *)
val is_bswap : Temu_trace.instruction -> bool

(** [is_arithmetic_raw itype] returns whether the given instruction mnemonic 
  [itype] corresponds to an arithmetic instruction. 
  E.g., add/adc/inc/dec/sub/sbb/mul/imul/div/idiv/lea *)
val is_arithmetic_raw : insnMnemonic -> bool

(** [is_arithmetic insn] returns whether the given instruction [insn] is an 
  arithmetic instruction *)
val is_arithmetic : Temu_trace.instruction -> bool

(** [is_bitwise_raw itype] returns whether the given instruction mnemonic 
  [itype] corresponds to a bitwise instruction. 
  E.g., or/and/xor/not/neg/rol/ror/rcl/rcr/shl/shr/sal/sar *)
val is_bitwise_raw : insnMnemonic -> bool

(** [is_bitwise insn] returns whether the given instruction [insn] is a 
  bitwise instruction *)
val is_bitwise : Temu_trace.instruction -> bool

(** [sets_eflags_raw itype] returns whether the given instruction mnemonic 
  [itype] corresponds to an instruction that modifies the EFLAGS register. *)
val sets_eflags_raw : insnMnemonic -> bool

(** [sets_eflags insn] returns whether the given instruction [insn] is an 
  instruction that modifies the EFLAGS register. *)
val sets_eflags : Temu_trace.instruction -> bool

(** [generates_constant_output insn] returns whether the given instruction 
  [insn] is an instruction that generates constant output. 
  E.g., XOR/SUB with identical operands, AND with zero immediate, 
    OR with minus one immediate *)
val generates_constant_output : Temu_trace.instruction -> bool

(** [is_mov_esp_ebp insn] returns whether the given instruction [insn] is a 
  "mov %esp, %ebp" instruction (commonly used in function prologues) *)
val is_mov_esp_ebp : Temu_trace.instruction -> bool

(** [is_nop insn] returns whether the given instruction [insn] is a 
  no-op, that is an instruction with no side effects that can 
  be used for paddin/alignment by assemblers. E.g., mov %esi, %esi *) 
val is_nop : Temu_trace.instruction -> bool

