(*
  Copyright (C) BitBlaze, 2009-2010. 

#Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.

#THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*)
(**
   A driver to operate on execution traces generated by TEMU. 
   
*)
open Vine_util
open Insn_classifier
open Callstack
module Trace = Temu_trace

type cmdlineargs_t = {
    mutable trace_filename : string; 
    mutable fmap_filename : string;
    mutable calls_filename : string;
    mutable modlist_filename : string;
    mutable funlist_filename : string;
    mutable first : int64; 
    mutable last : int64;
    mutable count : bool;  
    mutable output : string; 
    mutable taintpair : (int32*int32) option;
    mutable tainted_only : bool;
    mutable print_header : bool;
    mutable verbose : bool;
    mutable reduced_flog : bool;
    mutable tid : int32 option;
    mutable pid : int option;
    mutable function_stack_start_size : int;
    mutable createindex : bool; 
    mutable eiptrace : bool;
    mutable extra_modules_l : (int64*int64*string) list;
  }

(* A Reader context *)
type reader_context = {
  (* Trace interface *)
  trace_iface : Temu_trace.trace_interface;
  (* Process name *)
  proc_name : string;
  (* Start counter *)
  start_ctr : int64;
  (* Stop counter *)
  stop_ctr : int64;
  (********************** OUTPUT CHANNELS ************************)
  (* Output channel to write disas. Currently fixed to stdout *)
  oc : out_channel;
  (* Output channel for instructions (used for cutting a piece of trace) *)
  insn_oc_opt : unit IO.output option;
  (* Output channel for call log *)
  call_log_oc_opt : out_channel option;
  (********************** FUNCTION INFORMATION *******************)
  (* A callstack context *)
  cs_ctx_opt : (unit callstack_context) option;
  (* Module base map *)
  module_base_map_opt : (string,(Libasmir.address_t*int64)) Hashtbl.t option;
  (* Generate call log *)
  output_call_log : bool;
  (* Generate reduced function log *)
  output_reduced_call_log : bool;
  (* Keep track of which modules appear in the trace *)
  track_mod_seen : bool;
  mod_seen_map_opt : (string,bool) Hashtbl.t option;
  (* Keep track of which functions appear in the trace *)
  track_fun_seen : bool;
  fun_seen_map_opt : (string*string,bool) Hashtbl.t option;
  (* Current function is first level call *)
  mutable first_level_call : bool;
  (********************** WHAT TO PROCESS ************************)
  (* Process tainted instruction only *)
  process_tainted_only : bool;
  (* Process instructions with given taint pair only *)
  process_taint_pair_opt : (int32 * int32) option;
  (* Process instructions with given process identifier only *)
  process_pid_opt : int option;
  (* Process instructions with given thread identifier only *)
  process_tid_opt : int32 option;
  (********************** WHAT TO PRINT ************************)
  (* Print call/ret information *)
  print_call_ret_info : bool;
  (* Verbose. Print additional instruction information *)
  print_verbose : bool;
  (* Print EIP only instead of instruction information *)
  print_eip_only : bool;
  (* Print instruction counter in front of instruction's information *)
  print_insn_ctr : bool;
}

(* Get process name from interface *)
let get_proc_name trace_iface = 
  try (List.hd trace_iface#processes)#name 
  with _ -> "unknown"

(* Get PID of traced process from interface *)
let get_traced_pid trace_iface = 
  try (List.hd trace_iface#processes)#pid
  with _ -> (
    Printf.fprintf stderr "No process in trace\n";
    exit 1
  )

(* Create a reader context *)
let create_reader_ctx start_ctr stop_ctr
      insn_oc_opt function_map_opt module_map_opt call_log_oc_opt
      fun_seen_map_opt mod_seen_map_opt output_reduced_call_log
      process_tainted_only process_taint_pair_opt process_pid_opt 
      process_tid_opt print_eip_only print_insn_ctr verbose 
      trace_iface =

  let cs_ctx_opt = 
    match function_map_opt, module_map_opt with
      | Some(fun_map),Some(mod_map) -> 
          let cs_ctx = create_callstack_ctx fun_map mod_map in
          Some(cs_ctx)
      | _,_ -> None
  in

  {
    trace_iface = trace_iface;
    proc_name =  get_proc_name trace_iface;
    start_ctr = start_ctr;
    stop_ctr = stop_ctr;
    oc = stdout;
    insn_oc_opt = insn_oc_opt;
    call_log_oc_opt = call_log_oc_opt;
    cs_ctx_opt = cs_ctx_opt;
    module_base_map_opt = module_map_opt;
    output_call_log = Option.is_some call_log_oc_opt;
    output_reduced_call_log = output_reduced_call_log;
    track_mod_seen = Option.is_some mod_seen_map_opt;
    mod_seen_map_opt = mod_seen_map_opt;
    track_fun_seen = Option.is_some fun_seen_map_opt;
    fun_seen_map_opt = fun_seen_map_opt;
    first_level_call = false;
    process_tainted_only = process_tainted_only;
    process_taint_pair_opt = process_taint_pair_opt;
    process_pid_opt = process_pid_opt;
    process_tid_opt = process_tid_opt;
    print_call_ret_info = Option.is_some function_map_opt;
    print_verbose = verbose;
    print_eip_only = print_eip_only;
    print_insn_ctr = print_insn_ctr;
  }

    
(* Format operand address as a string *)
let opaddr_str op = 
  let default_str = Printf.sprintf "0x%08Lx" op#opaddr in
  match op#optype with 
    | Trace.TRegister -> (
	match op#opaddr with
	  (* segment registers *)
	  | 100L -> "es"
	  | 101L -> "cs"
	  | 102L-> "ss"
	  | 103L -> "ds"
	  | 104L -> "fs"
	  | 105L -> "gs"
	  (*address-modifier dependent registers *)
	  | 108L -> default_str
	  | 109L -> default_str
	  | 110L -> default_str
	  | 111L -> default_str
	  | 112L -> default_str
	  | 113L -> default_str
	  | 114L -> default_str
	  | 115L -> default_str
	  (* 8-bit registers *)
	  | 116L -> "al"
	  | 117L -> "cl"
	  | 118L -> "dl"
	  | 119L -> "bl"
	  | 120L -> "ah"
	  | 121L -> "ch"
	  | 122L -> "dh"
	  | 123L -> "bh"
	  (* 16-bit registers *)
	  | 124L -> "ax"
	  | 125L -> "cx"
	  | 126L -> "dx"
	  | 127L -> "bx"
	  | 128L -> "sp"
	  | 129L -> "bp"
	  | 130L -> "si"
	  | 131L -> "di"
	  (* 32-bit registers *)
	  | 132L -> "eax"
	  | 133L -> "ecx"
	  | 134L -> "edx"
	  | 135L -> "ebx"
	  | 136L -> "esp"
	  | 137L -> "ebp"
	  | 138L -> "esi"
	  | 139L -> "edi"
	  (* ??? *)
	  | 150L 
          | _ -> default_str
      )
    | Trace.TFloatRegister -> (
        match op#opaddr with
          (* Float registers *)
          | 188L -> "st0"
          | 189L -> "st1"
          | 190L -> "st2"
          | 191L -> "st3"
          | 192L -> "st4"
          | 193L -> "st5"
          | 194L -> "st6"
          | 195L -> "st7"
          | _ -> default_str
      )
    | Trace.TMMXRegister -> (
        match op#opaddr with
          (* MMX registers *)
          | 164L -> "mm0"
          | 165L -> "mm1"
          | 166L -> "mm2"
          | 167L -> "mm3"
          | 168L -> "mm4"
          | 169L -> "mm5"
          | 170L -> "mm6"
          | 171L -> "mm7"
          | _ -> default_str
      )
   | Trace.TXMMRegister -> (
        match op#opaddr with
          (* XMM registers *)
          | 172L -> "xmm0"
          | 173L -> "xmm1"
          | 174L -> "xmm2"
          | 175L -> "xmm3"
          | 176L -> "xmm4"
          | 177L -> "xmm5"
          | 178L -> "xmm6"
          | 179L -> "xmm7"
          | 180L -> "xmm8"
          | 181L -> "xmm9"
          | 182L -> "xmm10"
          | 183L -> "xmm11"
          | 184L -> "xmm12"
          | 185L -> "xmm13"
          | 186L -> "xmm14"
          | 187L -> "xmm15"
          | _ -> default_str
      )
    | Trace.TFloatControlRegister -> (
        match op#opaddr with
          (* Float control registers *)
          | 196L -> "fpuc"
          | 197L -> "fpus"
          | 198L -> "fputag"
          | _ -> default_str
      )
    | _ -> default_str


(* Format operand type as a string *)
let optype_str = function
    | Trace.TRegister -> "R"
    | Trace.TMemLoc -> "M"
    | Trace.TImmediate -> "I"
    | Trace.TJump -> "J"
    | Trace.TMemAddress -> "A"
    | Trace.TFloatControlRegister -> "R"
    | Trace.TFloatRegister -> "R"
    | Trace.TMMXRegister -> "R"
    | Trace.TXMMRegister -> "R"
    | _ -> ""

(* Format operand access as a string *)
let opaccess_str = function
  | Trace.A_RW -> "(RW)"
  | Trace.A_R -> "(R)"
  | Trace.A_W -> "(W)"
  | Trace.A_RCW -> "(RCW)"
  | Trace.A_CW -> "(CW)"
  | Trace.A_CRW -> "(CRW)"
  | Trace.A_CR -> "(CR)"
  | _ -> ""
  
(* Priority functions used when sortening operand list *)
let opaccess_priority = function
  | Trace.A_R -> 0
  | Trace.A_CR -> 1
  | Trace.A_CRW -> 2
  | Trace.A_RCW -> 3
  | Trace.A_RW -> 4
  | Trace.A_W -> 5
  | Trace.A_CW -> 6
  | _ -> 7

let optype_priority = function
  | Trace.TImmediate -> 0
  | Trace.TJump -> 1
  | Trace.TRegister -> 2
  | Trace.TMMXRegister -> 3
  | Trace.TXMMRegister -> 4
  | Trace.TFloatRegister -> 5
  | Trace.TFloatControlRegister -> 6
  | Trace.TMemAddress -> 7
  | Trace.TMemLoc -> 8
  | Trace.TNone -> 9


(* Operand comparison, used to order operands similarly to older traces *)
let operand_compare op1 op2 =
  let p1 = opaccess_priority op1#opaccess in
  let p2 = opaccess_priority op2#opaccess in
  let cmp1 = Pervasives.compare p1 p2 in
  if (cmp1 <> 0) then cmp1
  else (
    let pt1 = optype_priority op1#optype in 
    let pt2 = optype_priority op2#optype in 
    Pervasives.compare pt1 pt2
  )

(* Store module that instruction belongs to into mod_seen_map *)
let process_insn_module mod_base_map_opt mod_seen_map_opt inst = 
  match (mod_base_map_opt,mod_seen_map_opt) with 
    | (Some(mod_base_map),Some(mod_seen_map)) -> 
      let modname = find_module mod_base_map inst#address in
      Hashtbl.replace mod_seen_map modname true
    | _ -> ()

(* Print map with modules seen in trace *)
let print_mod_seen_map filename map_opt =
  match map_opt with 
    | Some(map) -> 
      let mod_oc = open_out filename in
      let process_mod modname _ l = 
	if (modname <> "unknown") 
	  then modname :: l
	  else l
      in
      let mod_seen_l = Hashtbl.fold process_mod map [] in 
      let sorted_mod_seen_l = List.sort (Pervasives.compare) mod_seen_l in
      List.iter (fun s -> Printf.fprintf mod_oc "%s\n" s) sorted_mod_seen_l;
      close_out mod_oc
    | None -> ()

(* Print function seen map *)
let print_funseen_map filename map_opt = 
  match map_opt with
    | Some(map) ->
      let oc = open_out filename in
      let keys = get_hash_keys map in
      let keys = List.sort (Pervasives.compare) keys in
      let print_pair prev_pair curr_pair = 
	if (curr_pair <> prev_pair) then (
	  let (mod_name,fun_name) = curr_pair in
	  Printf.fprintf oc "%s::%s\n" mod_name fun_name
	);
	curr_pair
      in
      let _ = List.fold_left print_pair ("","") keys in
      close_out oc
    | None -> ()

(* Given an operand, return a string containing the operands info, 
read to print *)
let format_operand_basic ?(sep="\t") opr =
  let format_base_operand opr = 
    Printf.sprintf "%s%s@%s[0x%08lx][%d]%s" sep (optype_str opr#optype)
      (opaddr_str opr) (opr#opvalue) opr#oplen (opaccess_str opr#opaccess)
  in
  let format_mmx_operand opr =
    let opval64 =
      match opr#opcontent with
        | Trace.Val64(val64) -> val64
        | _ -> 0L
    in
    Printf.sprintf "%s%s@%s[0x%016Lx][%d]%s" sep (optype_str opr#optype)
      (opaddr_str opr) opval64 opr#oplen (opaccess_str opr#opaccess)
  in
  let format_xmm_operand opr =
    let (h_val64,l_val64) =
      match opr#opcontent with
        | Trace.Val128(h64,l64) -> (h64,l64)
        | _ -> (0L,0L)
    in
    Printf.sprintf "%s%s@%s[0x%016Lx%016Lx][%d]%s" sep (optype_str opr#optype)
      (opaddr_str opr) h_val64 l_val64 opr#oplen (opaccess_str opr#opaccess)
  in
  let format_float_operand opr =
    let (exponent,significand) =
      match opr#opcontent with
        | Trace.ValFloat(e,s) -> (e,s)
        | _ -> (0,0L)
    in
    Printf.sprintf "%s%s@%s[0x%04x%016Lx][%d]%s" sep (optype_str opr#optype)
      (opaddr_str opr) exponent significand opr#oplen
      (opaccess_str opr#opaccess)
  in
  let format_taint_info opr_t =
    let taint_source i =
      if (Int64.logand opr_t#taintflag (Int64.of_int (1 lsl i))) <> 0L then
        Printf.sprintf "(%lu, %lu) " opr_t#origin.(i) opr_t#offset.(i)
      else
        "()"
    in
      match opr_t#taintflag with
          0L -> Printf.sprintf "%sT0" sep
        | _ -> (Printf.sprintf "%sT1 {%Lu " sep opr_t#taintflag)
            ^(taint_source 0) ^ (taint_source 1) ^
              (taint_source 2) ^ (taint_source 3) ^ "}"
  in
  match (opr#optype) with
      Trace.TRegister
    | Trace.TMemLoc
    | Trace.TImmediate
    | Trace.TJump
    | Trace.TMemAddress ->
	(format_base_operand opr) ^ (format_taint_info opr)
    | Trace.TFloatControlRegister ->
        (format_base_operand opr)
    | Trace.TFloatRegister ->
        (format_float_operand opr)
    | Trace.TMMXRegister ->
        (format_mmx_operand opr)
    | Trace.TXMMRegister ->
        (format_xmm_operand opr)
    | Trace.TNone -> ""

(* Same as format_operand_basic, but selects operand given inst and index *)
let format_operand inst index = 
  try (
    let opr = inst#operand.(index) in
    format_operand_basic opr
  )
  with _ -> ""
	  

(* Print trace header *)
let print_header oc ti = 
  Printf.fprintf oc "Trace version: %d\n" ti#version;
  Printf.fprintf oc "Number of instructions: %Ld\n" ti#num_instructions;
  let procs = ti#processes in
    let print_process proc = 
      Printf.fprintf oc "Process: %s PID: %d\n" proc#name proc#pid;
      let print_module curr_mod =
	Printf.fprintf oc "\t Module: %s @ 0x%08Lx Size: %Ld\n"
	  curr_mod#name curr_mod#base curr_mod#size
      in
	List.iter print_module proc#modules
    in
    let () = List.iter print_process procs in
    flush oc

(* Print an instruction *) 
let print_inst ?(print_ctr=false) ?(verbose=false) 
      first_ctr insn insn_ctr = 
  let _ = if ((Int64.sub insn_ctr first_ctr) > 0L) then print_newline () in
  (* Print instruction counter *)
  let () = 
    if print_ctr 
      then (Printf.printf "(%08Ld)" insn_ctr)
      else ()
  in
  (* Print address. Note that flushing here is necessary *)
  let () = Printf.printf "%Lx:\t%!" insn#address in
  (* Print disassembly *)
  let () = Asmir.print_i386_rawbytes insn#address insn#rawbytes in
  (* Get list of operands in instruction *)
  let op_l = List.rev (Array.to_list insn#operand) in
  (* Order list of operands using special comparison function on access *)
  let op_l = List.sort operand_compare op_l in
  (* Print list of operands *)
  let print_op op = 
    let argstr = format_operand_basic op in
    Printf.printf "%s" argstr;
  in
  let () = List.iter print_op op_l in

  (* If verbose, print additional information *)
  if verbose then (
    Printf.printf " ESP: %s " (format_operand_basic insn#esp);
    Printf.printf "NUM_OP: %d " insn#num_operands;
    Printf.printf "PID: %d " insn#pid;
    Printf.printf "TID: %ld " insn#thread_id;
    Printf.printf "TP: %s " (Trace.taintprop_str insn#tp);
    Printf.printf "EFLAGS: 0x%08lX " insn#eflags;
    Printf.printf "CC_OP: 0x%08lX " insn#cc_op;
    Printf.printf "DF: 0x%08lX " insn#df;
    (* Get instruction size *)
    let num_raw_bytes = 
      let is = insn#inst_size in
      if (is > 0) then is else 15
    in
    (* Print rawbytes *)
    Printf.printf "RAW: 0x";
    for i = 0 to num_raw_bytes - 1 do
      Printf.printf "%02x" (Char.code(insn#rawbytes.(i))) 
    done;
    (* Print memory addressing registers if any *)
(*    Printf.printf " MEMREGS: "; *)
    let num_ops = Array.length insn#operand in
    for i = 0 to (num_ops - 1) do
      try (
	let tmp_op = insn#operand.(i) in
	match tmp_op#optype with 
	| Trace.TMemLoc
	| Trace.TMemAddress -> 
	  (* Print segment, base, index *)
	  let memreg_fmt = 
	    ( (format_operand_basic insn#memregs.(i).(0)) ^ 
	      (format_operand_basic insn#memregs.(i).(1)) ^ 
	      (format_operand_basic insn#memregs.(i).(2))) 
	  in
	  if (memreg_fmt <> "") then (
	    (* Add displacement *)
	    let disp_fmt = 
	      try if (insn#memregs.(i).(5)#optype <> Trace.TNone) then (
		Printf.sprintf " disp: 0x%lx" insn#memregs.(i).(5)#opvalue
	      )
	      else "" with Invalid_argument "index out of bounds" -> ""
	    in
	    (* Add scale *)
	    let base_fmt = 
	      try if (insn#memregs.(i).(6)#optype <> Trace.TNone) then (
		Printf.sprintf " scale: %ld" insn#memregs.(i).(6)#opvalue
	      )
	      else "" with Invalid_argument "index out of bounds" -> ""
	    in
	    let segdesc_fmt =
	      try if (insn#memregs.(i).(3)#optype <> Trace.TNone) then (
		" seg descrs: " ^
		  (format_operand_basic insn#memregs.(i).(3)) ^
		  (format_operand_basic insn#memregs.(i).(4)))
	      else "" with Invalid_argument "index out of bounds" -> ""
	    in
	    Printf.printf " MEMREGS(%d): %s%s%s%s " i
	      memreg_fmt disp_fmt base_fmt segdesc_fmt;
	  )
	| _ -> ()
      )
      with _ -> ()
    done;
  )

(* Print function information *)
let print_fun_info ctx insn insn_ctr = 
  let cs_ctx = 
    match ctx.cs_ctx_opt with
      | Some(cs_ctx) -> cs_ctx
      | None -> failwith "Cannot print function information"
  in
  (* Get thread information *)
  let thread_info = Hashtbl.find cs_ctx.function_stack_map insn#thread_id in
  let (fun_info,fun_stats) = Stack.top thread_info.thread_fun_stack in
(*
  Printf.printf ("%06Ld %ld Is_entry: %s Is_exit: %s " ^^ 
    "Found_call: %s Found_ret: %s Num_pop: %d\n%!")
    insn_ctr insn#thread_id 
    (string_of_bool thread_info.thread_is_entry_inst)
    (string_of_bool thread_info.thread_is_exit_inst)
    (string_of_bool thread_info.thread_found_call)
    (string_of_bool thread_info.thread_found_ret)
    (List.length thread_info.thread_pop_l);
*)
  (* Found entry point *)
  (* if ((thread_info.thread_is_entry_inst) && (not (is_ret insn))) then ( *)
  if thread_info.thread_is_entry_inst then (
    (* Print function name *)
    let stack_len = Stack.length thread_info.thread_fun_stack in
    Printf.printf "\t%s::%s (%d)"
      fun_info.fi_modname fun_info.fi_funname stack_len;
    (* add_fun_to_funmap ctx.fun_seen_map_opt fun_info.fi_modname 
      fun_info.fi_funname insn_ctr; *)

    (* Write to call log *)
    let () =
      match ctx.call_log_oc_opt with
        | Some(call_oc) ->
          if ((not ctx.output_reduced_call_log) || ctx.first_level_call)
          then
            Printf.fprintf call_oc "%06Ld %ld CALL %s::%s (%d)\n"
              insn_ctr insn#thread_id fun_info.fi_modname fun_info.fi_funname
              stack_len
        | None -> ()
    in
    ctx.first_level_call <- false;
  )
  (* If entry point, print function name *)
  else if (thread_info.thread_is_exit_inst) then (
    let (fun_info,_) = 
      try List.hd thread_info.thread_pop_l 
      with _ -> (fun_info,fun_stats)
    in
    (* Print function name *)
    let stack_len = 
      let sl = Stack.length thread_info.thread_fun_stack in
      sl + (List.length thread_info.thread_pop_l)
    in
    Printf.printf "\t%s::%s (%d)" 
      fun_info.fi_modname fun_info.fi_funname stack_len;

    (* Write to call log *)
    match ctx.call_log_oc_opt with
      | Some(call_oc) -> (
          if (not ctx.output_reduced_call_log) then (
            Printf.fprintf call_oc "%06Ld %ld RET %s::%s (%d)\n"
              insn_ctr insn#thread_id fun_info.fi_modname fun_info.fi_funname 
              stack_len;
          )
          else if (thread_info.thread_ret_was_spurious) then (
            if (fun_info.fi_modname = ctx.proc_name) then (
              let (mod_name,fun_name) = 
                try (
                  let (mod_name,fun_name,_) =
                    Hashtbl.find cs_ctx.function_map insn#address
                  in
                  (mod_name,fun_name)
                )
                with Not_found -> (
                  let eip_str = Printf.sprintf "%Lx" insn#address in
                  (find_module cs_ctx.module_base_map insn#address, 
                    Printf.sprintf "sub_%s" (String.uppercase eip_str))
                )
              in
              Printf.fprintf call_oc "%06Ld %ld SPURIOUS_RET %s::%s\n"
                insn_ctr insn#thread_id mod_name fun_name
            )
          )
        )
      | None -> ()
  )
  else if (thread_info.thread_found_call) then (
    let modname = find_module cs_ctx.module_base_map insn#address in
    if (modname = ctx.proc_name) then (ctx.first_level_call <- true);
  )


(* Write instruction to output stream *)
let output_inst insn someoc =
    match someoc with
	Some(oc) -> insn#serialize oc
      | None -> ()

(* Check if instruction needs to be handled *)
let check_cond ctx insn =
  let checktaint insn (origin,offset) = 
    let op_l = (Array.to_list insn#operand) in
    let operand_uses_taintpair op = 
      let origin_l = Array.to_list op#origin in
      let offset_l = Array.to_list op#offset in
      let check_pair flag byte_origin byte_offset = 
        ((byte_origin = origin) && (byte_offset = offset)) || flag
      in
      List.fold_left2 check_pair false origin_l offset_l
    in
    List.exists operand_uses_taintpair op_l
  in
  let is_right_taintpair = 
    match ctx.process_taint_pair_opt with
      | None -> true
      | Some(origin,offset) -> checktaint insn (origin,offset)
  in
  let is_right_process = 
    match ctx.process_pid_opt with
      | None -> true
      | Some(pid) -> pid = insn#pid
  in
  let is_right_thread = 
    match ctx.process_tid_opt with
      | None -> true
      | Some(tid) -> tid = insn#thread_id
  in
  if (is_right_taintpair && is_right_process && is_right_thread) &&
      ((not ctx.process_tainted_only) || (is_tainted_instruction insn))
  then
    true
  else 
    false

(* Process instruction in trace *)
let tfold_print_insn ctx insn insn_ctr = 
  (* Update callstack information *)
  let () = 
    match ctx.cs_ctx_opt with 
      | Some(cs_ctx) -> 
          ignore(tfold_add_fun_to_callstack_ctx cs_ctx insn insn_ctr)
      | None -> ()
  in

  if (check_cond ctx insn)
  then (
    (* Write instruction in binary form to output file if requested *)
    output_inst insn ctx.insn_oc_opt;

    (* Add module to module seen map *)
    if ctx.track_mod_seen then (
      process_insn_module ctx.module_base_map_opt 
        ctx.mod_seen_map_opt insn
    );

    (* Write instruction information *)
    if ctx.print_eip_only then Printf.printf "%08Lx\n" insn#address
    else (
      (* Printf function call/ret info if requested *)
      if ctx.print_call_ret_info then print_fun_info ctx insn insn_ctr;

      (* Print instruction info *)
      print_inst ~print_ctr:ctx.print_insn_ctr ~verbose:ctx.print_verbose
        ctx.start_ctr insn insn_ctr
    )
  );
  ctx


let parsecmdline = 
  let cmdlineargs = {
      trace_filename = "";
      fmap_filename = "";
      calls_filename = "";
      modlist_filename = "";
      funlist_filename = "";
      first = 1L;
      last = Int64.max_int;
      count = false; 
      output = "";
      taintpair = None;
      tainted_only = false;
      print_header = false;
      verbose = false;
      reduced_flog = false;
      pid = None;
      tid = None;
      function_stack_start_size = 100;
      createindex = false; 
      eiptrace = false;
      extra_modules_l = [];
    }
  in
  let set_taintoffset s = 
    let offset = Int32.of_string s in
    let taintpair = 
      match cmdlineargs.taintpair with 
        | None -> Some(Int32.minus_one,offset)
        | Some(origin,_) -> Some(origin,offset)
    in
    cmdlineargs.taintpair <- taintpair
  in
  let set_taintorigin s = 
    let origin = Int64.to_int32 (Int64.of_string s) in
    let taintpair =
      match cmdlineargs.taintpair with
        | None -> Some(origin,Int32.minus_one)
        | Some(_,offset) -> Some(origin,offset)
    in
    cmdlineargs.taintpair <- taintpair
  in
  let add_module s = 
    let tl = Str.split (Str.regexp ":") s in
    try (
      let start_addr = Int64.of_string (List.hd tl) in
      let size = Int64.of_string (List.nth tl 1) in
      let name = List.nth tl 2 in
      cmdlineargs.extra_modules_l <- 
	(start_addr,size,name) :: cmdlineargs.extra_modules_l
    )
    with _ -> (
      Printf.fprintf stderr "Invalid module info: %s\n" s;
      exit 1
    )
  in
  let arg_spec = [
      ("-count", 
      Arg.Unit (fun s -> cmdlineargs.count <- true),
      "<> Display line number"); 
      
      ("-createindex",
      Arg.Unit (fun s -> cmdlineargs.createindex <- true),
      "<> Create an index file for the trace. Outputs trace.idx file"); 

      ("-eip",
      Arg.Unit (fun s -> cmdlineargs.eiptrace <- true),
      "<> Print only EIP of instruction");

      ("-emod",
      Arg.String add_module,
      ("<adddr>:<size>:<name> Add a module to the module list " ^
        "using given base address, module size and module name"));

      ("-first", 
      Arg.String (fun s -> cmdlineargs.first <- Int64.of_string s),
      "<int> Line number of first instruction"); 

      ("-fmap",
      Arg.String (fun s -> cmdlineargs.fmap_filename <- s),
      "<string> Name of input function map. Used to print function names " ^ 
	"at call/ret instructions");

      ("-flog",
      Arg.String (fun s -> cmdlineargs.calls_filename <- s),
      "<string> Name of output file to receive summary of call/ret in trace." ^ 
	" Requires -fmap option");  

      ("-flogr",
      Arg.String (fun s -> cmdlineargs.calls_filename <- s; 
	cmdlineargs.reduced_flog <- true),
      "<string> Same as -fmap but prints only calls from main module");  

      ("-funlist",
      Arg.String (fun s -> cmdlineargs.funlist_filename <- s),
      "<string> Print list of known functions seen in trace to given file");  

      ("-funstacksize",
      Arg.Int (fun x -> cmdlineargs.function_stack_start_size <- x),
      "<int> Set function stack start size to this value");

      ("-header",
      Arg.Unit (fun s -> cmdlineargs.print_header <- true),
      "<> Print trace header information and exit"); 

      ("-last", 
      Arg.String (fun s -> cmdlineargs.last <- Int64.of_string s),
      "<int> Line number of last instruction"); 

      ("-modlist",
      Arg.String (fun s -> cmdlineargs.modlist_filename <- s),
      "<string> Print list of modules seen in trace to given file");  

      ("-o",
      Arg.String (fun s -> cmdlineargs.output <- s),
      "<string> Output file name");

      ("-pid",
      Arg.Int (fun i -> cmdlineargs.pid <- Some(i)),
      "<int> Process only instructions with this process identifier");  

      ("-taintedonly", 
      Arg.Unit (fun s -> cmdlineargs.tainted_only <-true),
      "<> Process only tainted instructions"); 
      
      ("-taintpair", 
      Arg.Tuple ([Arg.String set_taintorigin; Arg.String set_taintoffset]),
      "<int32> <int32> Process only tainted instructions operating on this " ^ 
	"origin (int32) + offset (int) pair");

      ("-tid",
      Arg.Int (fun i -> cmdlineargs.tid <- Some(Int32.of_int i)),
      "<int> Process only instructions with this thread identifier");  

      ("-trace", 
      Arg.String (fun s -> cmdlineargs.trace_filename <- s),
      "<string> Name of input trace file"); 

      ("-v",
      Arg.Unit (fun () -> cmdlineargs.verbose <- true),
      "<> Verbose. Prints more info per instruction");

    ] 
  in 
  let () =
    Arg.parse arg_spec (fun s -> ()) 
      "Usage: trace_reader [options] -trace <trace_file>"
  in
    cmdlineargs
;;

(* Parse the command line *)
let args = parsecmdline in

(* Open the trace *)
let tracename = 
  if (args.trace_filename <> "")
  then
    args.trace_filename
  else (
    Printf.fprintf stderr "Need to specify trace file";
    exit 1
  )
in
let trace_iface = 
  try Trace.open_trace tracename 
  with _ -> (
    Printf.fprintf stderr "Could not open trace: %s" tracename;
    exit 1
  )
in

(* If requested, print header and exit *)
if (args.print_header) then (
  print_header stdout trace_iface;
  Trace.close_trace trace_iface;
  exit 0
);

(* If creating index, loop over all instructions and exit *)
if args.createindex then (
  let index_oc = open_out_bin (tracename^".idx") in
  let index_oc_io = IO.output_channel index_oc in
  let rec handle_inst trace2 =
    let current_offset = trace2#current_offset () in
    let inst_o =
      try
	Some(trace2#read_instruction)
      with
	| IO.No_more_input
	| Trace.Found_trailer -> None
    in match inst_o with
	None -> ()
      | Some(inst) -> 
	  let () = IO.write_i64 index_oc_io current_offset in
	  if ((args.last <> 0L) && (trace2#insn_counter > args.last))
        then ()
        else handle_inst trace2
  in
  let () = handle_inst trace_iface in
  let () = IO.close_out index_oc_io in
  let () = Trace.close_trace trace_iface in
  exit 0
);

(* If TID given but no PID given, set PID to be the default process *)
let () = 
  match args.pid, args.tid with
    | None, Some(tid) -> (
	let traced_pid = get_traced_pid trace_iface in
	Printf.fprintf stderr 
	  "Warning: No PID provided. Assuming TID from traced process %d\n"
	  traced_pid;
	args.pid <- Some(traced_pid)
      )
    | _, _ -> ()
in

(* Initialize output stream for instructions (writes header) *)
let insn_oc_opt =
  if (args.output <> "") then (
    let insn_oc = open_out args.output in
    let insn_oc_io = IO.output_channel insn_oc in
    trace_iface#write_header insn_oc_io;
    Some(insn_oc_io)
  )
  else None
in

(* Initialize function information *)
let (function_map_opt, module_map_opt, call_log_oc_opt,fun_seen_map_opt) = 
  if (args.fmap_filename <> "") then (
    (* Load function map *)
    let fun_map = read_function_map args.fmap_filename in
    (* Read start address for modules *)
    let mod_map = read_module_addresses trace_iface in
    (* Add user provided modules to module map *)
    let add_module (base,size,name) =
      Hashtbl.add mod_map name (base,size)
    in
    List.iter add_module args.extra_modules_l;
    (* Initialize function log ouput stream if needed *)
    let call_log_oc_opt =
      if (args.calls_filename <> "") then (
        let call_log_oc = open_out args.calls_filename in
        Some(call_log_oc)
      )
      else None
    in
    (* Initialize function seen map if needed *)
    let fun_seen_map = Hashtbl.create 40 in
    Some(fun_map),Some(mod_map),call_log_oc_opt,Some(fun_seen_map)
  )
  else None,None,None,None
in

(* Initialize module/function list file if needed *)
let (mod_seen_map_opt,module_map_opt) = 
  let mod_seen_map_opt = 
    if (args.modlist_filename <> "") then Some(Hashtbl.create 20)
    else None 
  in
  let mod_map_opt = 
    if (args.fmap_filename <> "") then module_map_opt
    else if (args.modlist_filename <> "") then 
      Some(read_module_addresses trace_iface)
    else None
  in
  (mod_seen_map_opt,mod_map_opt)
in

(* Create a reader context *)
let ctx =
  create_reader_ctx args.first args.last
      insn_oc_opt function_map_opt module_map_opt call_log_oc_opt
      fun_seen_map_opt mod_seen_map_opt args.reduced_flog
      args.tainted_only args.taintpair args.pid args.tid
      args.eiptrace args.count args.verbose 
      trace_iface
in

(* Iterate over instructions in trace *)
let ctx =
  Temu_trace.trace_fold ~start_ctr:args.first ~stop_ctr:args.last
    tfold_print_insn ctx trace_iface
in

(* Post_processing *)
let _ = 
  if (args.modlist_filename <> "") then (
    print_mod_seen_map args.modlist_filename ctx.mod_seen_map_opt
  );
  if (args.funlist_filename <> "") then (
    print_funseen_map args.funlist_filename ctx.fun_seen_map_opt
  )
in

(* Exit cleanly by closing file descriptors *)
let clean_exit = 
  (match insn_oc_opt with
    None -> ()
    | Some(oc) -> IO.close_out oc
  );
  (match call_log_oc_opt with
    None -> ()
    | Some(oc) -> close_out oc
  );
  print_newline (); 
  Trace.close_trace trace_iface; 
in
clean_exit

